#!/usr/bin/env python3

import asyncio
import concurrent.futures
import logging
from argparse import Namespace
from configparser import ConfigParser
from json import JSONDecodeError, load
from pathlib import Path
from shutil import rmtree
from typing import Awaitable, List
from urllib.parse import urlparse

from packaging.utils import canonicalize_name

from bandersnatch.verify import get_latest_json

logger = logging.getLogger(__name__)  # pylint: disable=C0103


def delete_path(blob_path: Path, dry_run: bool = False) -> int:
    if dry_run:
        logger.info(f" rm {blob_path}")
        return 0

    if not blob_path.exists():
        logger.debug(f"{blob_path} does not exist. Skipping")
        return 0

    try:
        if blob_path.is_dir():
            rmtree(blob_path)
        else:
            blob_path.unlink()
    except FileNotFoundError:
        # Due to using threads in executors we sometimes have a
        # race condition if canonicalize_name == passed in name
        pass
    except OSError:
        logger.exception(f"Unable to delete {blob_path}")
        return 1

    return 0


async def delete_packages(config: ConfigParser, args: Namespace) -> int:
    loop = asyncio.get_event_loop()
    workers = args.workers or config.getint("mirror", "workers")
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=workers)
    mirror_base_path = Path(config.get("mirror", "directory"))
    web_base_path = mirror_base_path / "web"
    json_base_path = web_base_path / "json"
    pypi_base_path = web_base_path / "pypi"
    simple_path = web_base_path / "simple"

    delete_coros: List[Awaitable] = []
    for package in args.pypi_packages:
        canon_name = canonicalize_name(package)
        need_nc_paths = canon_name != package
        json_full_path = json_base_path / canon_name
        json_full_path_nc = json_base_path / package if need_nc_paths else None
        legacy_json_path = pypi_base_path / canon_name
        logger.debug(f"Looking up {canon_name} metadata @ {json_full_path}")

        if not json_full_path.exists():
            if args.dry_run:
                logger.error(
                    f"Skipping {json_full_path} as dry run and no JSON file exists"
                )
                continue

            logger.error(f"{json_full_path} does not exist. Pulling from PyPI")
            await get_latest_json(json_full_path, config, executor, False)
            if not json_full_path.exists():
                logger.error(f"Unable to HTTP get JSON for {json_full_path}")
                continue

        with json_full_path.open("r") as jfp:
            try:
                package_data = load(jfp)
            except JSONDecodeError:
                logger.exception(f"Skipping {canon_name} @ {json_full_path}")
                continue

        for _release, blobs in package_data["releases"].items():
            for blob in blobs:
                url_parts = urlparse(blob["url"])
                blob_path = web_base_path / url_parts.path[1:]
                delete_coros.append(
                    loop.run_in_executor(executor, delete_path, blob_path, args.dry_run)
                )

        # Attempt to delete json, normal simple path + hash simple path
        package_simple_path = simple_path / canon_name
        package_simple_path_nc = simple_path / package if need_nc_paths else None
        package_hash_path = simple_path / canon_name[0] / canon_name
        package_hash_path_nc = (
            simple_path / canon_name[0] / package if need_nc_paths else None
        )
        # Try cleanup non canon name if they differ
        for package_path in (
            json_full_path,
            legacy_json_path,
            package_simple_path,
            package_simple_path_nc,
            package_hash_path,
            package_hash_path_nc,
            json_full_path_nc,
        ):
            if not package_path:
                continue

            delete_coros.append(
                loop.run_in_executor(executor, delete_path, package_path, args.dry_run)
            )

    if args.dry_run:
        logger.info("-- bandersnatch delete DRY RUN --")
    if delete_coros:
        logger.info(f"Attempting to remove {len(delete_coros)} files")
        return sum(await asyncio.gather(*delete_coros))
    return 0



import argparse
import asyncio
import logging
import logging.config
import shutil
import sys
from pathlib import Path
from tempfile import gettempdir
from typing import Any

import bandersnatch.configuration
import bandersnatch.delete
import bandersnatch.log
import bandersnatch.mirror
import bandersnatch.verify

logger = logging.getLogger(__name__)  # pylint: disable=C0103


# TODO: Workout why argparse.ArgumentParser causes type errors
def _delete_parser(subparsers: Any) -> None:
    d = subparsers.add_parser(
        "delete",
        help=(
            "Consulte metadata (locally or remotely) and delete "
            + "entire pacakge artifacts."
        ),
    )
    d.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Do not download or delete files",
    )
    d.add_argument(
        "--workers",
        type=int,
        default=0,
        help="# of parallel iops [Defaults to bandersnatch.conf]",
    )
    d.add_argument("pypi_packages", nargs="*")
    d.set_defaults(op="delete")


def _mirror_parser(subparsers: Any) -> None:
    m = subparsers.add_parser(
        "mirror",
        help="Performs a one-time synchronization with the PyPI master server.",
    )
    m.add_argument(
        "--force-check",
        action="store_true",
        default=False,
        help=(
            "Force bandersnatch to reset the PyPI serial (move serial file to /tmp) to "
            + "perform a full sync"
        ),
    )
    m.set_defaults(op="mirror")


def _verify_parser(subparsers: Any) -> None:
    v = subparsers.add_parser(
        "verify", help="Read in Metadata and check package file validity"
    )
    v.add_argument(
        "--delete",
        action="store_true",
        default=False,
        help="Enable deletion of packages not active",
    )
    v.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Do not download or delete files",
    )
    v.add_argument(
        "--json-update",
        action="store_true",
        default=False,
        help="Enable updating JSON from PyPI",
    )
    v.add_argument(
        "--workers",
        type=int,
        default=0,
        help="# of parallel iops [Defaults to bandersnatch.conf]",
    )
    v.set_defaults(op="verify")


def main() -> int:
    parser = argparse.ArgumentParser(description="PyPI PEP 381 mirroring client.")
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {bandersnatch.__version__}"
    )
    parser.add_argument(
        "-c",
        "--config",
        default="/etc/bandersnatch.conf",
        help="use configuration file (default: %(default)s)",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Turn on extra logging (DEBUG level)",
    )

    subparsers = parser.add_subparsers()
    _delete_parser(subparsers)
    _mirror_parser(subparsers)
    _verify_parser(subparsers)

    if len(sys.argv) < 2:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    bandersnatch.log.setup_logging(args)

    # Prepare default config file if needed.
    config_path = Path(args.config)
    if not config_path.exists():
        logger.warning(f"Config file '{args.config}' missing, creating default config.")
        logger.warning("Please review the config file, then run 'bandersnatch' again.")

        default_config_path = Path(__file__).parent / "default.conf"
        try:
            shutil.copy(default_config_path, args.config)
        except IOError as e:
            logger.error(f"Could not create config file: {e}")
        return 1

    config = bandersnatch.configuration.BandersnatchConfig(
        config_file=args.config
    ).config

    if config.has_option("mirror", "log-config"):
        logging.config.fileConfig(str(Path(config.get("mirror", "log-config"))))

    if args.op in ("delete", "verify"):
        loop = asyncio.get_event_loop()
        op_coro = (
            bandersnatch.delete.delete_packages
            if args.op == "delete"
            else bandersnatch.verify.metadata_verify
        )
        try:
            return loop.run_until_complete(op_coro(config, args))
        finally:
            loop.close()
    else:
        if args.force_check:
            status_file = Path(config.get("mirror", "directory")) / "status"
            if status_file.exists():
                tmp_status_file = Path(gettempdir()) / "status"
                try:
                    shutil.move(status_file, tmp_status_file)
                    logger.debug(
                        "Force bandersnatch to check everything against the master PyPI"
                        + f" - status file moved to {tmp_status_file}"
                    )
                except OSError as e:
                    logger.error(
                        f"Could not move status file ({status_file} to "
                        + f" {tmp_status_file}): {e}"
                    )
            else:
                logger.info(
                    f"No status file to move ({status_file}) - Full sync will occur"
                )

        return bandersnatch.mirror.mirror(config)


if __name__ == "__main__":
    exit(main())



"""
Module containing classes to access the bandersnatch configuration file
"""
import logging
import warnings
from configparser import ConfigParser
from typing import Any, Dict, NamedTuple, Optional, Type

import pkg_resources

logger = logging.getLogger("bandersnatch")


class DeprecatedKey(NamedTuple):
    old_section: str
    old_key: str
    new_section: str
    new_key: str
    deprecated_version: str


class Singleton(type):  # pragma: no cover
    _instances: Dict["Singleton", Type] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Type:
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class BandersnatchConfig(metaclass=Singleton):
    # Ensure we only show the deprecations once
    SHOWN_DEPRECATIONS = False
    DEPRECATED_KEYS = {
        # "friendly_name": "DeprecatedKey",
        "Enabling Plugins": DeprecatedKey(
            "blacklist", "plugins", "plugins", "enabled", "4.0.0"
        )
    }

    def __init__(self, config_file: Optional[str] = None) -> None:
        """
        Bandersnatch configuration class singleton

        This class is a singleton that parses the configuration once at the
        start time.

        Parameters
        ==========
        config_file: str, optional
            Path to the configuration file to use
        """
        self.default_config_file = pkg_resources.resource_filename(
            "bandersnatch", "default.conf"
        )
        self.config_file = config_file
        self.load_configuration()
        self.check_for_deprecations()

    def check_for_deprecations(self) -> None:
        if self.SHOWN_DEPRECATIONS:
            return

        for friendly_name, dk in self.DEPRECATED_KEYS.items():
            err_msg = (
                f"{friendly_name} keys will move from {dk.old_section}:{dk.old_key} "
                + f"to {dk.new_section}:{dk.new_key} in version {dk.deprecated_version}"
                + f" - Documentation @ https://bandersnatch.readthedocs.io/"
            )
            warnings.warn(err_msg, DeprecationWarning, stacklevel=2)
            logger.warning(err_msg)

        self.SHOWN_DEPRECATIONS = True

    def load_configuration(self) -> None:
        """
        Read the configuration from a configuration file
        """
        config_file = self.default_config_file
        if self.config_file:
            config_file = self.config_file
        self.config = ConfigParser()
        self.config.read(config_file)

        # Copy deprecated keys to the new keys if they exist
        if "blacklist" in self.config and "plugins" in self.config["blacklist"]:
            self.config["plugins"] = {}
            self.config["plugins"]["enabled"] = self.config["blacklist"]["plugins"]



import logging
from operator import itemgetter

from packaging.version import parse

from bandersnatch.filter import FilterReleasePlugin

logger = logging.getLogger("bandersnatch")


class LatestReleaseFilter(FilterReleasePlugin):
    """
    Plugin to download only latest releases
    """

    name = "latest_release"
    keep = 0  # by default, keep 'em all

    def initialize_plugin(self):
        """
        Initialize the plugin reading patterns from the config.
        """
        if self.keep:
            return

        try:
            self.keep = int(self.configuration["latest_release"]["keep"])
        except KeyError:
            return
        except ValueError:
            return
        if self.keep > 0:
            logger.info(f"Initialized latest releases plugin with keep={self.keep}")

    def filter(self, info, releases):
        """
        Keep the latest releases
        """

        if self.keep == 0:
            return

        versions = list(releases.keys())
        before = len(versions)

        if before <= self.keep:
            # not enough releases: do nothing
            return

        versions_pair = map(lambda v: (parse(v), v), versions)
        latest = sorted(versions_pair)[-self.keep :]  # noqa: E203
        latest = list(map(itemgetter(1), latest))

        current_version = info.get("version")
        if current_version and (current_version not in latest):
            # never remove the stable/official version
            latest[0] = current_version

        logger.debug(f"old {versions}")
        logger.debug(f"new {latest}")

        after = len(latest)
        latest = set(latest)
        for version in list(releases.keys()):
            if version not in latest:
                del releases[version]

        logger.debug(f"{self.name}: releases removed: {before - after}")



import logging
from typing import List

from bandersnatch.filter import FilterReleasePlugin

logger = logging.getLogger("bandersnatch")


class ExcludePlatformFilter(FilterReleasePlugin):
    """
    Filters releases based on regex patters defined by the user.
    """

    name = "exclude_platform"

    _patterns: List[str] = []
    _packagetypes: List[str] = []

    def initialize_plugin(self):
        """
        Initialize the plugin reading patterns from the config.
        """
        if self._patterns or self._packagetypes:
            logger.debug(
                "Skipping initalization of Exclude Platform plugin. "
                + "Already initialized"
            )
            return

        try:
            tags = self.configuration["blacklist"]["platforms"].split()
        except KeyError:
            logger.error(f"Plugin {self.name}: missing platforms= setting")
            return

        for platform in tags:
            lplatform = platform.lower()

            if lplatform in ("windows", "win"):
                # PEP 425
                # see also setuptools/package_index.py
                self._patterns.extend([".win32", "-win32", "win_amd64", "win-amd64"])
                # PEP 527
                self._packagetypes.extend(["bdist_msi", "bdist_wininst"])

            elif lplatform in ("macos", "macosx"):
                self._patterns.extend(["macosx_", "macosx-"])
                self._packagetypes.extend(["bdist_dmg"])

            elif lplatform in ("freebsd"):
                # concerns only very few files
                self._patterns.extend([".freebsd", "-freebsd"])

            elif lplatform in ("linux"):
                self._patterns.extend(
                    [
                        "linux-i686",  # PEP 425
                        "linux-x86_64",  # PEP 425
                        "linux_armv7l",  # https://github.com/pypa/warehouse/pull/2010
                        "linux_armv6l",  # https://github.com/pypa/warehouse/pull/2012
                        "manylinux1_",  # PEP 513
                        "manylinux2010_",  # PEP 571
                    ]
                )
                self._packagetypes.extend(["bdist_rpm"])

        logger.info(f"Initialized {self.name} plugin with {self._patterns!r}")

    def filter(self, info, releases):
        """
        Remove files from `releases` that match any pattern.
        """

        # Make a copy of releases keys
        # as we may delete packages during iteration
        removed = 0
        versions = list(releases.keys())
        for version in versions:
            new_files = []
            for file_desc in releases[version]:
                if self._check_match(file_desc):
                    removed += 1
                else:
                    new_files.append(file_desc)
            if len(new_files) == 0:
                del releases[version]
            else:
                releases[version] = new_files
        logger.debug(f"{self.name}: filenames removed: {removed}")

    def _check_match(self, file_desc) -> bool:
        """
        Check if a release version matches any of the specificed patterns.

        Parameters
        ==========
        name: file_desc
            file description entry

        Returns
        =======
        bool:
            True if it matches, False otherwise.
        """

        # source dist: never filter out
        pt = file_desc.get("packagetype")
        if pt == "sdist":
            return False

        # Windows installer
        if pt in self._packagetypes:
            return True

        fn = file_desc["filename"]
        for i in self._patterns:
            if i in fn:
                return True

        return False



import logging
import re

from bandersnatch.filter import FilterReleasePlugin

logger = logging.getLogger("bandersnatch")


class PreReleaseFilter(FilterReleasePlugin):
    """
    Filters releases considered pre-releases.
    """

    name = "prerelease_release"
    PRERELEASE_PATTERNS = (
        r".+rc\d+$",
        r".+a(lpha)?\d+$",
        r".+b(eta)?\d+$",
        r".+dev\d+$",
    )
    patterns = None

    def initialize_plugin(self):
        """
        Initialize the plugin reading patterns from the config.
        """
        if not self.patterns:
            self.patterns = [
                re.compile(pattern_string)
                for pattern_string in self.PRERELEASE_PATTERNS
            ]
            logger.info(f"Initialized prerelease plugin with {self.patterns}")

    def filter(self, info, releases):
        """
        Remove all release versions that match any of the specificed patterns.
        """
        for version in list(releases.keys()):
            if any(pattern.match(version) for pattern in self.patterns):
                del releases[version]
